
/*
	HTTP "UI" for displaying the current SoundPlay playlist on a 
	remote webbrowser.
	Copyright 1999 Marco Nelissen (marcone@xs4all.nl)

	Permission is hereby granted to use this code for creating other
	SoundPlay-plugins.
*/

#include <Application.h>
#include <Roster.h>
#include <AppFileInfo.h>
#include <Resources.h>
#include <UTF8.h>
#include <stdio.h>
#include <string>
#include <signal.h>

#include "HttpUI.h"

const char *PLUGINID="sp-http-playlist";

void *getplugin(void **data,const char *name, const char *, uint32, plugin_info *info, SoundPlayBufferedReader *reader);
void destroyplugin(void*,void *_loader);
status_t show(void*);
void hide(void*);
BView *configure(BMessage *config);
void setconfig(void*,BMessage *config);
void getconfig(void*,BMessage *config);

// =======================================================================
// standard plugin stuff
static plugin_descriptor plugin={
	PLUGIN_DESCRIPTOR_MAGIC,PLUGIN_DESCRIPTOR_VERSION,
	PLUGINID,2, PLUGIN_IS_INTERFACE,

	"HTTP interface",
	"By Marco Nelissen.\n\n"
	"Displays the playlist as a webpage. "
	"Connect your browser to port 8080\n",
	NULL,
	&configure,
	&getplugin,
	&destroyplugin
};

static interface_plugin_ops plugin_ops={
	PLUGIN_INTERFACE_MAGIC, PLUGIN_INTERFACE_VERSION,
	show,
	hide,
	setconfig,
	getconfig
};

static plugin_descriptor *plugs[]={
	&plugin,0
};

static version_info libbe_version;
static bool sockets_are_files = false;

plugin_descriptor **get_plugin_list(void)
{
	BFile libbe("/boot/beos/system/lib/libbe.so",B_READ_ONLY); // let's hope they don't move it
	BAppFileInfo ai(&libbe);
	if(B_OK!=ai.GetVersionInfo(&libbe_version,B_APP_VERSION_KIND))
		libbe_version.short_info[0]=0;
	if(strcmp(libbe_version.short_info, "5.1")>=0)
		sockets_are_files = true;
	return plugs;
}

// =======================================================================
// actual implementation

int32 listenfunc(void *data);
int32 handlerfunc(void *data);

void *getplugin(void **data,const char *name, const char *, uint32, plugin_info *info, SoundPlayBufferedReader *reader)
{
	NetHandle *net=new NetHandle(info->controller);

	BFile file(info->ref,B_READ_ONLY);
	BResources resource;
	if(B_OK==resource.SetTo(&file))
	{
		const void *playbut=resource.LoadResource('bits', "playbutton", &net->playbuttonsize);
		if(playbut)
		{
			net->playbutton=(char*)malloc(net->playbuttonsize);
			memcpy(net->playbutton,playbut,net->playbuttonsize);
		}
		else
			net->playbutton=NULL;

		const void *downbut=resource.LoadResource('bits', "downloadbutton", &net->downloadbuttonsize);
		if(downbut)
		{
			net->downloadbutton=(char*)malloc(net->downloadbuttonsize);
			memcpy(net->downloadbutton,downbut,net->downloadbuttonsize);
		}
		else
			net->downloadbutton=NULL;

		const void *streambut=resource.LoadResource('bits', "streambutton", &net->streambuttonsize);
		if(streambut)
		{
			net->streambutton=(char*)malloc(net->streambuttonsize);
			memcpy(net->streambutton,streambut,net->streambuttonsize);
		}
		else
			net->streambutton=NULL;
	}

	*data=net;
	return &plugin_ops;
}

void destroyplugin(void *,void *data)
{
	NetHandle *net=(NetHandle*)data;
	
	net->endpoint->Close();
	if(net->listener>=0)
	{
		status_t dummy;
		send_signal(net->listener,SIGUSR1);
		wait_for_thread(net->listener,&dummy);
	}
	delete net->endpoint;
	delete net;
}

status_t show(void *data)
{
	NetHandle *net=(NetHandle*)data;

	net->endpoint=new BNetEndpoint(SOCK_STREAM);
	if(net->endpoint->InitCheck()==B_OK)
	{
		if(sockets_are_files)
			fcntl(net->endpoint->Socket(), F_SETFD, FD_CLOEXEC);

		status_t ret;
#if 0
		ret=net->endpoint->Bind(8080);	// forces a dialup connection
#elif 0
		ret=net->endpoint->Bind(BNetAddress("127.0.0.1",8080));	// doesn't force dialup, but only listens on 127.0.0.1, not the external IP
#elif 1
		ret=net->endpoint->Bind(BNetAddress(INADDR_ANY,8080));	// forces a dialup connection
#else
		// code stolen from PoorMan, yet it *still* forces a dialup (and PoorMan doesn't)
		struct sockaddr_in mine;
		mine.sin_family = AF_INET;
		mine.sin_port = htons(8080);
		mine.sin_addr.s_addr = 0;
		memset(mine.sin_zero, 0, sizeof(mine.sin_zero));
		ret=net->endpoint->Bind(BNetAddress(mine));
#endif
		if(ret!=B_OK)
			return B_ERROR;

		if(B_OK!=net->endpoint->Listen())
			return B_ERROR;
		net->listener=spawn_thread(listenfunc,"SP HTTP listener",B_NORMAL_PRIORITY,net);
		if(net->listener>=0)
		{
			resume_thread(net->listener);
			return B_OK;
		}
	}
	return B_ERROR;
}

int32 listenfunc(void *data)
{
	NetHandle *net=(NetHandle*)data;
	BNetEndpoint *remote;
	while((remote=net->endpoint->Accept())!=NULL)
	{
		if(B_OK==remote->InitCheck())
		{
			NetHandle *remotehandle=new NetHandle(net->ctrl);
			*remotehandle=*net;
			remotehandle->endpoint=remote;

			// spawn a thread to do the actual communication.
			// I should be keeping track of all the threads I create, but I'm too lazy
			thread_id worker=spawn_thread(handlerfunc,"SP HTTP worker",B_NORMAL_PRIORITY,remotehandle);
			if(worker>=0)
				resume_thread(worker);
		}
	}
	return B_OK;
}

string itoa(int value, char *format="%d")
{
	char _itoabuf[64];
	sprintf(_itoabuf,format,value);
	return string(_itoabuf);
}

string escaped_name(char *s)
{
	string escname;
	while(*s)
	{
		if(*s!=' ')
			escname+=*s;
		else
			escname+="%20";
		s++;
	}
	return escname;
}

void showplaylist(NetHandle *net)
{
	BNetBuffer netbuffer; // use a buffer so we don't need to keep SoundPlay locked during transfers

	if(B_OK!=netbuffer.InitCheck())
	{
		string header=	"HTTP/1.0 500 Internal Server Error\n"
						"Location: /\n\n";
		net->endpoint->Send(header.data(),header.length());
		return;
	}

	string header=
		"HTTP/1.0 200 OK\n"
		"Content-type: text/html\n"
		"Server: BeOS/SoundPlay\n\n"
		"<HTML>\n"
		"<HEAD>"
		"<META HTTP-EQUIV=\"Content-Type\" CONTENT=\"text/html; charset=UTF-8\">"
		"<TITLE>SoundPlay playlist</TITLE>"
		"</HEAD>"
		"<BODY BGCOLOR=\"#FFFFFF\">"
		"<meta http-equiv=refresh content=\"90\">"
		"<FONT SIZE=\"30\"><A HREF=\"http://www.xs4all.nl/~marcone/soundplay.html\">SoundPlay ";
	header += net->ctrl->VersionString();
	header +="</A></FONT><P>";
	header +="Running on BeOS ";
	header +=libbe_version.short_info;
	header+="<P><HR><P>\n";

	netbuffer.AppendData(header.data(),header.length());
	net->ctrl->Lock();
	bool hastracks=false;
	for(int t=0;;t++)
	{
		PlaylistPtr playlist=net->ctrl->PlaylistAt(t);
		if(!playlist->IsValid())
			break;
		hastracks=true;
		string playerline;
		playerline+="<FONT SIZE=\"+2\">Player ";
		playerline+=itoa(t+1);
		playerline+="</FONT><P>\n";
		playerline+="<TABLE BORDER=\"0\" CELLSPACING=\"1\">\n";
		netbuffer.AppendData(playerline.data(),playerline.length());
		bool hasfiles=false;
		playlist->Lock();
		int32 curfile=playlist->CurrentID();
		for(int idx=0;;idx++)
		{
			int32 id=playlist->IDForItemAt(idx);
			if(id<0)
				break;
			const char *path=playlist->PathForItem(id);

			entry_ref ref;
			status_t refstatus=get_ref_for_path(path,&ref);
			hasfiles=true;
			string filename="";
			filename+="<TR>";
			if(curfile==id)
				filename+="<TD BGCOLOR=\"#c0c0c0\">&nbsp;";
			else
				filename+="<TD BGCOLOR=\"#e0e0e0\">&nbsp;";
			filename+=itoa(idx+1,"%d");
			filename+="</TD>";
			if(curfile==id)
				filename+="<TD BGCOLOR=\"#c0c0c0\">&nbsp;";
			else
				filename+="<TD BGCOLOR=\"#e0e0e0\">&nbsp;";

			// first encode from utf8 to iso-8859-1
			unsigned char encodebuffer[B_FILE_NAME_LENGTH];
			char namesourcebuffer[B_PATH_NAME_LENGTH];

			if(refstatus==B_OK)
				strcpy(namesourcebuffer,ref.name);
			else if(path)
				strcpy(namesourcebuffer,path);
			else
				strcpy(namesourcebuffer,"<unnamed file>");

			int32 encodebufferlen=B_FILE_NAME_LENGTH;
			int32 sourcelen=strlen(namesourcebuffer);
			int32 state=0;
			convert_from_utf8(B_ISO1_CONVERSION,playlist->NameForItem(id),&sourcelen,(char*)encodebuffer,&encodebufferlen,&state);
			encodebuffer[encodebufferlen]='\0';
			// replace "special" character with their numeric encoding (&#nnn;)
			unsigned char *s=encodebuffer;
			while (*s)
			{
				// replacements to prevent filenames being interpreted as html-tags
				if(*s=='&')
					filename+="&amp;";
				else if(*s=='<')
					filename+="&lt;";
				else if(*s=='>')
					filename+="&gt;";

				// standard ascii stuff
				else if(*s<=127)
					filename+=*s;

				// some common accented characters, feel free to add more
				else if(*s==224)
					filename+="&agrave;";
				else if(*s==225)
					filename+="&aacute;";
				else if(*s==228)
					filename+="&auml;";
				else if(*s==232)
					filename+="&egrave;";
				else if(*s==233)
					filename+="&eacute;";
				else if(*s==234)
					filename+="&ecirc;";
				else if(*s==235)
					filename+="&euml;";
				else if(*s==242)
					filename+="&ograve;";
				else if(*s==243)
					filename+="&oacute;";
				else if(*s==246)
					filename+="&ouml;";
				else if(*s==249)
					filename+="&ugrave;";
				else if(*s==250)
					filename+="&uacute;";
				else if(*s==252)
					filename+="&uuml;";

				s++;
			}
			filename+="&nbsp;</TD>";
			
			if(net->showfileinfo)
			{
				BNode node(&ref);

				char attribute[100];
				if(curfile==id)
					filename+="<TD BGCOLOR=\"#c0c0c0\">&nbsp;";
				else
					filename+="<TD BGCOLOR=\"#e0e0e0\">&nbsp;";

				if(node.InitCheck())
				{
					ssize_t attrsize=node.ReadAttr("Audio:Length",B_STRING_TYPE, 0, attribute,99);
					if(attrsize>0)
					{
						attribute[attrsize]=0; // add termination just in case
						filename+=attribute;
					}
				}
				filename+="&nbsp;</TD>";

				if(curfile==id)
					filename+="<TD BGCOLOR=\"#c0c0c0\">&nbsp;";
				else
					filename+="<TD BGCOLOR=\"#e0e0e0\">&nbsp;";
				if(node.InitCheck())
				{
					ssize_t attrsize=node.ReadAttr("Audio:Bitrate",B_STRING_TYPE, 0, attribute,99);
					if(attrsize>0)
					{
						attribute[attrsize]=0; // add termination just in case
						filename+=attribute;
					}
				}
				filename+="&nbsp;</TD>";
			}

			if(net->playbuttonsenabled)
			{
				filename+="<TD><A HREF=\"/play/"+itoa(t)+"/"+itoa(id)+"\"><IMG SRC=\"playbutton.gif\" BORDER=\"0\" ALT=\"play\"></A></TD>";
			}
			if(net->downbuttonsenabled)
			{
				filename+="<TD><A HREF=\"/dlod/"+itoa(t)+"/"+itoa(id)+"/"+escaped_name(namesourcebuffer)+"\"><IMG SRC=\"downloadbutton.gif\" BORDER=\"0\" ALT=\"download\"></A></TD>";
			}
			if(net->streambuttonsenabled)
			{
				filename+="<TD><A HREF=\"/strm/"+itoa(t)+"/"+itoa(id)+"/playlist.pls\"><IMG SRC=\"streambutton.gif\" BORDER=\"0\" ALT=\"stream\"></A></TD>";
			}
			filename+="\n";
			netbuffer.AppendData(filename.data(),filename.length());
		}
		playlist->Unlock();
		if(!hasfiles)
		{
			char *err="No files\n";
			netbuffer.AppendData(err,strlen(err));
		}
		netbuffer.AppendData("</TABLE>",8);
		netbuffer.AppendData("<P>",3);
	}
	if(!hastracks)
	{
		char *err="<FONT SIZE=\"+2\">No players</FONT><P>";
		netbuffer.AppendData(err,strlen(err));
	}
	net->ctrl->Unlock();

	string footer=
		"<P>"
		"<HR>"
		"<FONT SIZE=\"-2\">"
		"Generated ";

	time_t now=time(NULL);
	footer+=ctime(&now);
	footer+=
		" by SoundPlay HTTP interface plugin"
		"</BODY>\n"
		"</HTML>\n";
	netbuffer.AppendData(footer.data(),footer.length());
	net->endpoint->Send(netbuffer);
}


int32 handlerfunc(void *data)
{
	NetHandle *net=(NetHandle*)data;
	char request[2048];
	char filename[64];
	int32 numread;
	bool parse_error=true;
	
	// read the client's request.
	numread=net->endpoint->Receive(request,sizeof(request)-1);
	request[numread]=0;
//	printf("request: %s\n",request);

	// parse request
	if(sscanf(request,"GET %63s HTTP",filename)==1)
	{
		if(strcmp(filename,"/")==0)
		{
			showplaylist(net);
			parse_error=false;
		}
		else if((strncmp(filename,"/play/",6)==0)&&(net->playbuttonsenabled))
		{
			int t,f;
			if(sscanf(filename,"/play/%d/%d",&t,&f)==2)
			{
				net->ctrl->Lock();
				PlaylistPtr playlist=net->ctrl->PlaylistAt(t);
				if(playlist->IsValid())
				{
					playlist->Lock();
					playlist->PlayFile(f);
					playlist->Unlock();
				}
				net->ctrl->Unlock();

				string header=	"HTTP/1.0 302 Moved Temporarily\n"
								"Location: /\n\n";
				net->endpoint->Send(header.data(),header.length());
				parse_error=false;
			}
		}
		else if((strncmp(filename,"/dlod/",6)==0)&&
				(net->downbuttonsenabled || net->streambuttonsenabled))
		{
			int t,id;
			if(sscanf(filename,"/dlod/%d/%d",&t,&id)==2)
			{
				entry_ref ref;
				net->ctrl->Lock();
				PlaylistPtr playlist=net->ctrl->PlaylistAt(t);
				if(playlist->IsValid())
				{
					playlist->Lock();
					const char *path=playlist->PathForItem(id);
					if(path)
						get_ref_for_path(path,&ref);
					playlist->Unlock();
				}
				net->ctrl->Unlock();
				BEntry entry(&ref);
				if(entry.Exists())
				{
					string response=
							"HTTP/1.0 200 OK\n"
							"Content-type: audio/x-mpeg\n"
							"Server: BeOS/SoundPlay\n\n";
					net->endpoint->Send(response.data(),response.length());

					BFile file(&ref,B_READ_ONLY);
					ssize_t filenumread;
					int32 filenumsent;
					
					char filebuf[16384];
					do {
						filenumread=file.Read(filebuf,16384);
						filenumsent=0;
						if(filenumread>0)
							filenumsent=net->endpoint->Send(filebuf,filenumread);
					} while((filenumread>0)&&(filenumsent==filenumread));
				}
				else
				{
					string error=
							"HTTP/1.0 200 OK\n"
							"Content-type: text/html\n"
							"Server: BeOS/SoundPlay\n\n"
							"<HTML>\n"
							"<HEAD>"
							"<META HTTP-EQUIV=\"Content-Type\" CONTENT=\"text/html; charset=UTF-8\">"
							"<TITLE>404 File not found</TITLE>"
							"</HEAD>\n"
							"<BODY BGCOLOR=\"#FFFFFF\">\n"
							"<FONT SIZE=\"+10\">404 File not found</FONT>\n"
							"</BODY>\n"
							"</HTML>\n";
					net->endpoint->Send(error.data(),error.length());
				}
				parse_error=false;
			}
		}
		else if((strncmp(filename,"/strm/",6)==0)&&(net->streambuttonsenabled))
		{
			int t,f;
			if(sscanf(filename,"/strm/%d/%d",&t,&f)==2)
			{
				char *host=strstr(request,"\nHost: ");
				if(host!=NULL)
				{
					host++;
					char hostname[201];
					if(sscanf(host,"Host: %200s",hostname)==1)
					{
						string header=
							"HTTP/1.0 200 OK\n"
							"Content-type: audio/x-scpls\n"
							"Server: BeOS/SoundPlay\n\n"
							"[playlist]\n"
							"File1=http://";
						header+=hostname;
						if(strstr(hostname,":")==NULL)
						{
							header+=":";
							header+=itoa(8080);
						}
						header+="/dlod/";
						header+=itoa(t);
						header+="/";
						header+=itoa(f);
						header+="/stream.mp3\n";
						header+="NumberOfEntries=1\n";
						net->endpoint->Send(header.data(),header.length());
						parse_error=false;
					}
				}
			}
		}
		else if(strcmp(filename,"/playbutton.gif")==0)
		{
			string response=
					"HTTP/1.0 200 OK\n"
					"Content-type: image/gif\n"
					"Server: BeOS/SoundPlay\n\n";
			net->endpoint->Send(response.data(),response.length());
			net->endpoint->Send(net->playbutton,net->playbuttonsize);
			parse_error=false;
		}
		else if(strcmp(filename,"/downloadbutton.gif")==0)
		{
			string response=
					"HTTP/1.0 200 OK\n"
					"Content-type: image/gif\n"
					"Server: BeOS/SoundPlay\n\n";
			net->endpoint->Send(response.data(),response.length());
			net->endpoint->Send(net->downloadbutton,net->downloadbuttonsize);
			parse_error=false;
		}
		else if(strcmp(filename,"/streambutton.gif")==0)
		{
			string response=
					"HTTP/1.0 200 OK\n"
					"Content-type: image/gif\n"
					"Server: BeOS/SoundPlay\n\n";
			net->endpoint->Send(response.data(),response.length());
			net->endpoint->Send(net->streambutton,net->streambuttonsize);
			parse_error=false;
		}
	}

	if(parse_error)
	{
		string error=
				"HTTP/1.0 400 ERROR\n"
				"Content-type: text/html\n"
				"Server: BeOS/SoundPlay\n\n"
				"<HTML>\n"
				"<HEAD>"
				"<META HTTP-EQUIV=\"Content-Type\" CONTENT=\"text/html; charset=UTF-8\">"
				"<TITLE>ERROR</TITLE>"
				"</HEAD>\n"
				"<BODY BGCOLOR=\"#FFFFFF\">\n"
				"<FONT SIZE=\"+10\">Request not understood</FONT>\n"
				"</BODY>\n"
				"</HTML>\n";
		net->endpoint->Send(error.data(),error.length());
	}

	net->endpoint->Close();
	delete net->endpoint;
	delete net;
	return B_OK;
}

void hide(void *data)
{
}

BView* configure(BMessage *config)
{
	return new ConfigView(config);
}

void setconfig(void *data,BMessage *config)
{
	NetHandle *net=(NetHandle*)data;
	net->showfileinfo=config->FindBool("info");
	net->playbuttonsenabled=config->FindBool("play");
	net->downbuttonsenabled=config->FindBool("down");
	net->streambuttonsenabled=config->FindBool("stream");
}

void getconfig(void *data,BMessage *config)
{
}



ConfigView::ConfigView(BMessage *_config)
	: VGroup((MView*)NULL)
{
	config=_config;

	// make default config
	if(!config->HasBool("info"))
		config->AddBool("info",false);
	if(!config->HasBool("play"))
		config->AddBool("play",false);
	if(!config->HasBool("down"))
		config->AddBool("down",false);
	if(!config->HasBool("stream"))
		config->AddBool("stream",false);

	AddChild(new MCheckBox("Show file info",new BMessage('info'),this,config->FindBool("info")));
	AddChild(new MCheckBox("Enable play-links",new BMessage('play'),this,config->FindBool("play")));
	AddChild(new MCheckBox("Enable download-links",new BMessage('down'),this,config->FindBool("down")));
	AddChild(new MCheckBox("Enable stream-links",new BMessage('strm'),this,config->FindBool("stream")));
}

ConfigView::~ConfigView()
{

}

void ConfigView::MessageReceived(BMessage *mes)
{
	switch(mes->what)
	{
		case 'info':
			config->ReplaceBool("info",mes->FindInt32("be:value"));
			break;
		case 'play':
			config->ReplaceBool("play",mes->FindInt32("be:value"));
			break;
		case 'down':
			config->ReplaceBool("down",mes->FindInt32("be:value"));
			break;
		case 'strm':
			config->ReplaceBool("stream",mes->FindInt32("be:value"));
			break;
	}
}

