/*
	'Loudness' plugin for SoundPlay.
	Copyright 1999 Marco Nelissen (marcone@xs4all.nl)

	This is an example of a filter plugin for SoundPlay.
	Permission is hereby granted to use this code for creating other
	SoundPlay-plugins.
*/


#include <Application.h>
#include <SupportDefs.h>
#include <TypeConstants.h>
#include <ByteOrder.h>
#include <string.h>
#include <stdio.h>
#include <MediaDefs.h>
#include <Alert.h>
#include <malloc.h>

#include "pluginproto.h"
#include "LoudnessPlugin.h"

// ===============================
// interface to the outside world

plugin_descriptor plugin={
	PLUGIN_DESCRIPTOR_MAGIC,PLUGIN_DESCRIPTOR_VERSION,
	"sp-loudness",1, PLUGIN_IS_FILTER,

	"Loudness",
	"By Marco Nelissen.\n\n"
	"Emphasises low and high frequencies.\n",
	aboutplugin,
	configureplugin,
	&getplugin,
	&destroyplugin
};

static plugin_descriptor *plugs[]={
	&plugin,
	0
};

plugin_descriptor **get_plugin_list(void) // this function gets exported
{
	return plugs;
}


// ===============================
// interface to a plugin-instance

filter_plugin_ops Loudness_ops={
	// first some bookkeeping stuff
	PLUGIN_FILTER_MAGIC, PLUGIN_FILTER_VERSION,

	// and the function pointers.
	NULL,	// FileChange
	Loudness_filter,
	Loudness_configure,
	NULL,
	NULL
};



/* ============================================== */

/* Digital filter designed by mkfilter/mkshape/gencode   A.J. Fisher 
   Command line: /www/usr/fisher/helpers/mkfilter -Bu -Bs -o 2 -a 1.8140589569e-02 1.8140589569e-01 -l */ 

#define NZEROS 4 
#define NPOLES 4 
#define GAIN   2.113652662e+00 

typedef float FLOAT;

class datastruct
{
public:
	FLOAT lxv[NZEROS+1];
	FLOAT lyv[NPOLES+1];
	FLOAT rxv[NZEROS+1];
	FLOAT ryv[NPOLES+1];
	datastruct()
	{
		for(int i=0;i<=NZEROS;i++)
			lxv[i]=rxv[i]=0;
		for(int i=0;i<=NPOLES;i++)
			lyv[i]=ryv[i]=0;
	}
};


void *getplugin(void **data,const char *, const char *, uint32, plugin_info *, SoundPlayBufferedReader *reader)
{
	*data=new datastruct;
	return &Loudness_ops;
}

void destroyplugin(void *,void *_data)
{
	datastruct *data=(datastruct*)_data;
	delete data;
}


// Normally you wouldn't implement the about function for such a simple
// message. It is done here for completeness.
static void aboutplugin()
{
	BAlert *alert=new BAlert("About Loudness plugin",
		"Loudness plugin.\n"
		"By Marco Nelissen.\n",
		"OK");
	alert->SetFeel(B_NORMAL_WINDOW_FEEL);
	alert->Go(NULL);
}

static BView* configureplugin(BMessage*)
{
	BAlert *alert=new BAlert("main config",
		"Main configuration goes here.",
		"OK");
	alert->SetFeel(B_NORMAL_WINDOW_FEEL);
	alert->Go(NULL);
	return NULL;
}



BView* Loudness_configure(void * /*data*/)
{
	BAlert *alert=new BAlert("runtime config",
		"Runtime configuration goes here.",
		"OK");
	alert->SetFeel(B_NORMAL_WINDOW_FEEL);
	alert->Go(NULL);
	return NULL;
}

unsigned long idum=0;

status_t Loudness_filter(void *_data, short *_buffer, int32 framecount, filter_info *)
{
	datastruct *data=(datastruct*)_data;
	int16 *input=(int16*)_buffer;

	FLOAT *lxv=data->lxv;
	FLOAT *lyv=data->lyv;
	FLOAT *rxv=data->rxv;
	FLOAT *ryv=data->ryv;
	
	for(int i=0; i < framecount;i++)
	{
		// add some semi-random noise to stabilize the filter;
		idum=1664525L*idum+1013904223L;

		// left channel
		int is=(*input+(idum&1))<<1;
		lxv[0] = lxv[1]; lxv[1] = lxv[2]; lxv[2] = lxv[3]; lxv[3] = lxv[4]; 
		lxv[4] = FLOAT(is) / GAIN; 
		lyv[0] = lyv[1]; lyv[1] = lyv[2]; lyv[2] = lyv[3]; lyv[3] = lyv[4]; 
		lyv[4] =  (lxv[0] + lxv[4]) -   3.7178296355 * (lxv[1] + lxv[3]) +   5.4555642997 * lxv[2] 
				+ ( -0.2463565519 * lyv[0]) + (  1.0584801114 * lyv[1]) 
				+ ( -2.2809800042 * lyv[2]) + (  2.4594390844 * lyv[3]);
		int output= (int(lyv[4])>>2)+*input;
		if(output<-32768)
			output=-32768;
		else if(output>32767)
			output=32767;
		*input++ = output;

		// right channel
		is=(*input+(idum&1))<<1;
		rxv[0] = rxv[1]; rxv[1] = rxv[2]; rxv[2] = rxv[3]; rxv[3] = rxv[4]; 
		rxv[4] = FLOAT(is) / GAIN;
		ryv[0] = ryv[1]; ryv[1] = ryv[2]; ryv[2] = ryv[3]; ryv[3] = ryv[4]; 
		ryv[4] =  (rxv[0] + rxv[4]) -   3.7178296355 * (rxv[1] + rxv[3]) +   5.4555642997 * rxv[2] 
				+ ( -0.2463565519 * ryv[0]) + (  1.0584801114 * ryv[1]) 
				+ ( -2.2809800042 * ryv[2]) + (  2.4594390844 * ryv[3]); 
		output= (int(ryv[4])>>2)+*input;
		if(output<-32768)
			output=-32768;
		else if(output>32767)
			output=32767;
		*input++ = output;

	}
	return B_OK;
}

int main()
{
	BApplication app("application/x-marcone-loudnessplugin");
	BAlert *alert=new BAlert("About Loudness filter",
		"Loudness filter plugin for SoundPlay.\n"
		"By Marco Nelissen.\n",
		"OK");
	alert->SetFeel(B_NORMAL_WINDOW_FEEL);
	alert->Go();
}




