/*
	MIDI plugin for SoundPlay.
	Copyright 1999 Marco Nelissen (marcone@xs4all.nl)

	This is an example of a non data producing plugin for SoundPlay.
	Permission is hereby granted to use this code for creating other
	SoundPlay-plugins.
*/

#include <Application.h>
#include <Alert.h>
#include <MidiSynthFile.h>
#include <MediaDefs.h>

#include "pluginproto.h"
#include "MidiPlugin.h"


/* ===========================================================
	Plugin-descriptor structure.
	This structure is used to gather information about a plugin,
	and	instantiate and destroy specific instances of a plugin.
*/
static plugin_descriptor plugin={
	PLUGIN_DESCRIPTOR_MAGIC,PLUGIN_DESCRIPTOR_VERSION,
	"sp-midi",1, PLUGIN_IS_DECODER,

	"BMidiSynth MIDI player",
	"By Marco Nelissen.",
	&aboutplugin,
	NULL, // configure
	&getplugin,
	&destroyplugin
};

/* ===========================================================
	Plugin-operations structure.
	This is used to actually perform operations on a plugin.
	It is returned by the getplugin() function from the 
	plugin_descriptor structure above.
*/
static decoder_plugin_ops ops={
	PLUGIN_DECODER_MAGIC,PLUGIN_DECODER_VERSION,

	NULL, // open
	NULL, // close
	info,
	NULL, // read, not implemented because this plugin doesn't produce sampledata
	play,
	stop,
	setvolume,
	setspeed,
	seek,
	position,
	NULL  // bufferamount
};


/* ===========================================================
	Return the list of plugins in this file.
*/
static plugin_descriptor *plugs[]={
	&plugin,
	0
};

plugin_descriptor **get_plugin_list(void)
{
	return plugs;
}

/* ===========================================================
	Return the list of filetypes supported by all the plugins
	in this file
*/
static supported_type supportedtypes[]=
{
	{"audio/x-midi",				"MIDI file",					"MIDI file",				"mid","midi" },
	{NULL,							NULL,							NULL,						NULL,NULL}
};

supported_type *get_supported_types(void)
{
	return supportedtypes;
}


/* ===========================================================
	Implementation of the plugin_descriptor functions
*/

/*
	There is a bug in BSynth::CountClients() in R4.5 that
	prevents us from getting the actual number of connected
	clients. We therefore keep our own counter, but this
	can interfere with other plugins that also use BSynth.
	Since there probably won't be any, this is OK.
*/
static int numclients=0;

static void *getplugin(void **data, const char *name, const char *header, uint32 size, plugin_info *, SoundPlayBufferedReader *reader)
{
	entry_ref ref;
	if(B_OK==get_ref_for_path(name,&ref))
	{
		printf("ref OK\n");
		BEntry entry(&ref);
		if(entry.Exists())
		{
			BMidiSynthFile *synth=new BMidiSynthFile();
			if(synth->LoadFile(&ref)==B_OK)
			{
				*data=(void*)synth;
				be_synth->Resume();
				numclients++;
				return &ops;
			}
			delete synth;
		}
	}

	if(be_synth && numclients==0)
		be_synth->Pause();
	return NULL;
}
#include <stdio.h>
static void destroyplugin(void *,void *_synth)
{
	BMidiSynthFile *synth=(BMidiSynthFile *)_synth;
	synth->Stop();
	synth->UnloadFile();
	delete synth;
	numclients--;
	if(numclients==0)
		be_synth->Pause();
}

// Normally you wouldn't implement the about function for such a simple
// message. It is done here for completeness.
static void aboutplugin()
{
	BAlert *alert=new BAlert("About MIDI player",
		"Native BeOS BMidiSynth player.\n"
		"By Marco Nelissen.\n",
		"OK");
	alert->SetFeel(B_NORMAL_WINDOW_FEEL);
	alert->Go(NULL);
}



/* ===========================================================
	Implementation of the decoder_plugin_ops functions
*/

static status_t info(void *_synth, file_info *info)
{
	BMidiSynthFile *synth=(BMidiSynthFile *)_synth;

// Fill in the file_info structure
// This plugin tries to provide a broad indication of how long 
// the file is going to play, but since BMidiSynthFile cannot
// accurately provide this, it also sets the flags to indicate
// that no elapsed/remaining time calculation should be done
// based on the framecount and Position()

	strcpy(info->name,"");
	strcpy(info->typedesc,"MIDI file");
	strcpy(info->mimetype,"audio/x-midi");
	info->samplerate=1000*synth->Tempo();
	info->bitrate=1*4*8;
	info->numchannels=2;
	info->granularity=1;					// in "ticks"
	info->framecount=synth->Duration();		// in "ticks"
	info->samplesize=2;
	info->byteorder=B_BIG_ENDIAN;
	info->sampleformat=B_LINEAR_SAMPLES;
	info->flags=PLUGIN_FILELENGTH_UNKNOWN|PLUGIN_NO_ELAPSEDTIME;
	
	return B_OK;
}

static status_t play(void *_synth)
{
	BMidiSynthFile *synth=(BMidiSynthFile *)_synth;
	if(synth->IsFinished())
	{
		synth->Start();
	}
	else
	{
		synth->Resume();
	}
	return B_OK;
}

static status_t stop(void *_synth)
{
	BMidiSynthFile *synth=(BMidiSynthFile *)_synth;
	synth->Pause();
	return B_OK;
}

static status_t setvolume(void *_synth, float vol)
{
	BMidiSynthFile *synth=(BMidiSynthFile *)_synth;
	synth->SetVolume(vol);
	return B_OK;
}

static status_t	seek(void *_synth, uint32 pos)
{
	BMidiSynthFile *synth=(BMidiSynthFile *)_synth;
	synth->Position(pos); // not a typo...
	return B_OK;
}

static uint32	position(void *_synth)
{
	BMidiSynthFile *synth=(BMidiSynthFile *)_synth;
	return synth->Seek(); // not a typo...
}

static status_t setspeed(void *_synth, float speed)
{
	BMidiSynthFile *synth=(BMidiSynthFile *)_synth;
	
	if(speed<0)
		speed=0;
	if(speed!=0)
		synth->Resume();
	synth->ScaleTempoBy(speed);
	return B_OK;
}


int main()
{
	BApplication app("application/x-marcone-midiplugin");
	BAlert *alert=new BAlert("About MIDI player",
		"MIDI plugin for SoundPlay.\n"
		"By Marco Nelissen.\n",
		"OK");
	alert->SetFeel(B_NORMAL_WINDOW_FEEL);
	alert->Go();
}

