
/*
	Spectrum Analyzer visualization plugin for SoundPlay.
	Copyright 2001 Marco Nelissen (marcone@xs4all.nl)

	Permission is hereby granted to use this code for creating other
	SoundPlay-plugins.
*/

#include <Application.h>
#include <SupportDefs.h>
#include <TypeConstants.h>
#include <ByteOrder.h>
#include <string.h>
#include <stdio.h>
#include <MediaDefs.h>
#include <Alert.h>
#include <malloc.h>
#include <Bitmap.h>
#include <StopWatch.h>
#include <Screen.h>
#include <View.h>
#include "pluginproto.h"
#include "SpectrumAnalyzer.h"

plugin_descriptor plugin={
	PLUGIN_DESCRIPTOR_MAGIC,PLUGIN_DESCRIPTOR_VERSION,
	"sp-SpectrumAnalyzer",1, PLUGIN_IS_VISUAL,

	"SpectrumAnalyzer",
	"By Marco Nelissen.\n\n",
	NULL, // about
	NULL, // configure
	&getspectrumplugin,
	&destroyspectrumplugin
};

void SetConfig(void*, BMessage*);
void GetConfig(void*, BMessage*);

filter_plugin_ops ops={
	// first some bookkeeping stuff
	PLUGIN_VISUAL_MAGIC, PLUGIN_VISUAL_VERSION,

	// and the function pointers.
	spectrumfilechange,
	spectrumfilter,
	NULL,
	SetConfig,
	GetConfig
};

static plugin_descriptor *plugs[]={
	&plugin,
	0
};

plugin_descriptor **get_plugin_list(void) // this function gets exported
{
	return plugs;
}


class SpectrumAnalyzerWindow: public BWindow
{
	private:
		BView *rView;
		BView *bView;
		BBitmap *bitmap;
		SoundPlayController *controller;
		bool keeprunning;
		uint16 peak[20];
		uint16 lastpeak[20];
		uint16 peakdelay[20];
		const char *name;

	public:
		SpectrumAnalyzerWindow(SoundPlayController *ctrl);
		virtual ~SpectrumAnalyzerWindow();
		virtual bool QuitRequested();
		virtual void MessageReceived(BMessage *mes);
		status_t Render(const short *buffer, int32 count);
};

void *getspectrumplugin(void **data,const char *, const char *, uint32, plugin_info *info, SoundPlayBufferedReader *reader)
{
	SpectrumAnalyzerWindow *win=new SpectrumAnalyzerWindow(info->controller);
	*data=win;
	return &ops;
}

void destroyspectrumplugin(void*,void *data)
{
	SpectrumAnalyzerWindow *win=(SpectrumAnalyzerWindow *)data;
	win->Lock();
	win->Quit();
}

void spectrumfilechange(void *data, const char *name, const char *path)
{
	SpectrumAnalyzerWindow *win=(SpectrumAnalyzerWindow *)data;
	BMessage msg('file');
	msg.AddString("name",name);
	win->PostMessage(&msg);
}

status_t spectrumfilter(void *data, short *buffer, int32 count, filter_info*)
{
	SpectrumAnalyzerWindow *win=(SpectrumAnalyzerWindow *)data;
	if(win->Lock())
	{
		status_t status=win->Render(buffer, count);
		win->Unlock();
		return status;
	}
	return B_ERROR;
}

void SetConfig(void *data, BMessage *config)
{
	SpectrumAnalyzerWindow *win=(SpectrumAnalyzerWindow *)data;
	BRect frame;
	if(B_OK==config->FindRect("frame",&frame))
	{
		win->Lock();
		win->MoveTo(frame.LeftTop());
		win->Unlock();
	}
}

void GetConfig(void *data, BMessage *config)
{
	SpectrumAnalyzerWindow *win=(SpectrumAnalyzerWindow *)data;
	win->Lock();
	config->MakeEmpty();
	config->AddRect("frame",win->Frame());
	win->Unlock();
}


// ==============================================================================
SpectrumAnalyzerWindow::SpectrumAnalyzerWindow(SoundPlayController *ctrl)
	: BWindow(BRect(200,100,199+244,99+128),"Spectrum Analyzer",B_TITLED_WINDOW,B_NOT_RESIZABLE|B_NOT_ZOOMABLE),
	name(NULL)
{
	controller=ctrl;
	controller->AddWindow(this);	// let SoundPlay handle some hotkeys for this window

	rView=new BView(BRect(0,0,256-1,128-1),"view",B_FOLLOW_ALL_SIDES, 0);
	rView->SetViewColor(B_TRANSPARENT_COLOR);
	rView->SetFontSize(20);
	AddChild(rView);

	bitmap=new BBitmap(rView->Bounds(),B_COLOR_8_BIT,true);
	bitmap->Lock();
	bView=new BView(rView->Bounds(),"view",B_FOLLOW_ALL_SIDES, 0);
	bitmap->AddChild(bView);
	bView->SetFontSize(9);
	bitmap->Unlock();

	for(int x=0;x<20;x++)
	{
		peak[x]=0;
		lastpeak[x]=0;
		peakdelay[x]=0;
	}
	keeprunning=true;
	Show();
}

SpectrumAnalyzerWindow::~SpectrumAnalyzerWindow()
{
	delete bitmap;
}

bool SpectrumAnalyzerWindow::QuitRequested()
{
	keeprunning=false;
	return false;
}


status_t SpectrumAnalyzerWindow::Render(const short *buf, int32 framecount)
{
	if(keeprunning)
	{
		bitmap->Lock();
		bView->SetHighColor(0,0,0,255);
		bView->SetLowColor(0,0,0,255);
		bView->FillRect(bView->Bounds());

		const uint16 *fft = controller->GetFFTForBuffer(buf,SP_MONO_CHANNEL);

		bView->SetHighColor(100,100,100,255);
		for(int i=0;i<8;i++)
			bView->StrokeLine(BPoint(0,125-i*15),BPoint(256,125-i*15));
		bView->SetHighColor(0,220,0,255);
		if(name)
			bView->DrawString(name,BPoint(3,13));

		for(int x=0;x<20;x++)
		{
			uint32 total=0;
			for(int i=0;i<3+x;i++)
				total+=*fft++;
			total/=4*512-x*90;
			if(total>110)
				total=110;

			if(total<lastpeak[x])
				total=(total+lastpeak[x]*3)/4;
			lastpeak[x]=total;

			if(total)
			{
				BRect bar(2+x*12,125-total,2+x*12+10,125);
				bView->SetHighColor(0,150,0,255);
				bView->StrokeRect(bar);
				bar.InsetBy(1,1);
				bView->SetHighColor(0,220,0,255);
				bView->FillRect(bar);
			}
			if(total>peak[x])
			{
				peak[x]=total;
				peakdelay[x]=20;
			}
			if(peakdelay[x])
				peakdelay[x]--;
			if(peakdelay[x]==0)
				if(peak[x])
					peak[x]--;
			bView->SetHighColor(255,0,0,255);
			bView->FillRect(BRect(2+x*12,123-peak[x],2+x*12+10,124-peak[x]));
		}
		bView->Sync();
		rView->DrawBitmapAsync(bitmap,bitmap->Bounds(),rView->Bounds());
		rView->Flush();
		rView->Sync();

		bitmap->Unlock();
		return B_OK;
	}
	return B_ERROR;
}


void SpectrumAnalyzerWindow::MessageReceived(BMessage *message)
{
	switch(message->what)
	{
		case 'file':
			{
				const char *oldname = name;
				const char *newname = message->FindString("name");
				name = strdup(newname?newname:"");
				free((void*)oldname);
			}
			break;
			
			
		default:
			BWindow::MessageReceived(message);
			break;
	}
}

